#include "helpers.c"
#include <stdio.h>

/*
RAM: rambus dynamic Random-Access Memory
REG: rdram REGisters
RCP: Reality CoProcessor
LEO: LEO registers (64DD)
IPL: Initial Program Load (64DD)
STA: STAtic ram
ROM: ROM (cartridge)
PIF: Program Information File
*/

/* enumerated memory segments */
enum SEGMENT{RAM, REG, RCP, LEO, IPL, STA, ROM, PIF};

/* lookup table for the purpose of translating n64 memory segments to our segments */
BYTE LUT[0x20]=
{/* 0x00           0x03 0x04 0x05 0x06      0x08                                    */
     RAM,   0,   0, REG, RCP, LEO, IPL,   0, STA,   0,   0,   0,   0,   0,   0,   0,
     
 /* 0x10                                                                       0x1F */
     ROM,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, PIF
};

WORD loop;

/* function for sign extension of immediate values */
WORD extend(HALF imm);
WORD extend(HALF imm)
{
     WORD extended;
     if(sign(imm)==0x8000)
     {
          extended=0xffff0000|imm;
     }
     else
     {
          extended=imm;
     }
     return extended;
}

/* function for reading 32 bits */
WORD read32(BYTE *mem[8], WORD ptr);
WORD read32(BYTE *mem[8], WORD ptr)
{
     BYTE temp=(ptr>>24)&0x1f;
     switch(LUT[temp])
     {
          case REG:
               ptr&=0x0fffff;
               break;
          case PIF:
               ptr&=0x3fffff;
               break;
          default:
               ptr&=0xffffff;
               break;
     }
     if(ptr==0x600010){return 0;}
     return (WORD)((mem[LUT[temp]][ptr]<<24)|(mem[LUT[temp]][ptr+1]<<16)|(mem[LUT[temp]][ptr+2]<<8)|mem[LUT[temp]][ptr+3]);
}

unsigned long long read64(BYTE *mem[8], WORD ptr)
{
     BYTE temp=(ptr>>24)&0x1f;
     switch(LUT[temp])
     {
          case REG:
               ptr&=0x0fffff;
               break;
          case PIF:
               ptr&=0x3fffff;
               break;
          default:
               ptr&=0xffffff;
               break;
     }
     return (unsigned long long)((mem[LUT[temp]][ptr]<<56)|(mem[LUT[temp]][ptr+1]<<48)|(mem[LUT[temp]][ptr+2]<<40)|(mem[LUT[temp]][ptr+3]<<32)|(mem[LUT[temp]][ptr+4]<<24)|(mem[LUT[temp]][ptr+5]<<16)|(mem[LUT[temp]][ptr+6]<<8)|mem[LUT[temp]][ptr+7]);
}

void write32(BYTE *mem[8], WORD ptr, WORD val);
void handle_pi(BYTE *mem[8], WORD ptr);

void handle_pi(BYTE *mem[8], WORD ptr)
{
     WORD dram_addr, cart_addr, rd_len, wr_len;
     dram_addr=read32(mem,0xa4600000);
     cart_addr=read32(mem,0xa4600004)&0xfffffff;
     rd_len=1+read32(mem,0xa4600008);
     wr_len=1+read32(mem,0xa460000c);
     if(ptr==0x600008)
     {
     }
     else if(ptr==0x60000c)
     {
          if(cart_addr==0x1000)
          {
               if(mem[PIF][0x7e6]==0x91)
               {
                    mem[RAM][0x3f1]=0x80;
               }
               else
               {
                    mem[RAM][0x319]=0x80;
               }
          }
          for(loop=0;loop<wr_len;loop+=1)
          {
               mem[RAM][dram_addr+loop]=mem[ROM][cart_addr+loop];
          }
     }
}

/* function for writing 32 bits */
void write32(BYTE *mem[8], WORD ptr, WORD val)
{
     BYTE handle=9;
     BYTE temp=(ptr>>24)&0x1f;
     switch(LUT[temp])
     {
          case REG:
               ptr&=0x0fffff;
               break;
          case RCP:
               handle=(ptr&0xf00000)>>20;
               ptr&=0xffffff;
               break;
          case PIF:
               ptr&=0x3fffff;
               break;
          default:
               ptr&=0xffffff;
               break;
     }
     mem[LUT[temp]][ptr]=(val>>24)&0xff;
     mem[LUT[temp]][ptr+1]=(val>>16)&0xff;
     mem[LUT[temp]][ptr+2]=(val>>8)&0xff;
     mem[LUT[temp]][ptr+3]=val&0xff;
     switch(handle)
     {
          case 6:
               handle_pi(mem,ptr);
               break;
     }
}

char pointers[1159][64]={0};
int where[1159],where2[1159];

void zelda(BYTE *mem[8])
{
    FILE *text;
    int addr,add,clue,loop2;
    WORD mips;
    addr=0xB830;
    for(loop=0;loop<1159;loop+=1)
    {
        add=0;
        while(mem[ROM][addr]!=0)
        {
            pointers[loop][add]=mem[ROM][addr];
            add+=1;
            addr+=1;
        }
        while(mem[ROM][addr]==0)
        {
            addr+=1;
        }
        where[loop]=read32(mem,0x100116A0+0x10*loop);
        where2[loop]=read32(mem,0x100116A4+0x10*loop);
    }
    return;
    scan:
    printf("Scan for? ");
    scanf("%x",&clue);
    if(clue==0){return;}
    if(clue==1)
    {
        text=fopen("zeloot_rcp.txt","w");
        for(clue=0;clue<3381;clue+=1)
        {
            for(loop2=1606;loop2<1679;loop2+=1)
            {
            for(loop=offsets[clue];loop<offsets[clue+1];loop+=4)
            {
                mips=read32(mem,0x10000000+loop);
                if(op(mips)==JAL)
                {
                    if(target(mips)+0xA771A0==offsets[loop2])
                    {
                        fprintf(text,"%02d|%06X %-42s %s\n",loop2-1606,loop,symbols[clue],symbols[loop2]);
                        break;
                    }
                }
            }
            }
        }
        fprintf(text,"\n");
        for(clue=29;clue<498;clue+=1)
        {
            addr=0;
            for(loop2=1606;loop2<1679;loop2+=1)
            {
            for(loop=where[clue];loop<where2[clue];loop+=4)
            {
                mips=read32(mem,0x10000000+loop);
                if(op(mips)==JAL)
                {
                    if(target(mips)+0xA771A0==offsets[loop2])
                    {
                        addr+=1;
                        fprintf(text,"%02d|%06X %-25s %s\n",loop2-1606,loop,pointers[clue],symbols[loop2]);
                        break;
                    }
                }
            }
            }
            if(addr>0){fprintf(text,"count: %d\n",addr);}
        }
        fclose(text);
        goto pause;
    }
    if(clue==2)
    {
        //text=fopen("zeloot_rcp.txt","w");
        for(clue=0;clue<3381;clue+=1)
        {
            for(loop2=288;loop2<370;loop2+=1)
            {
            for(loop=offsets[clue];loop<offsets[clue+1];loop+=4)
            {
                mips=read32(mem,0x10000000+loop);
                if(op(mips)==JAL)
                {
                    if(target(mips)+0xA771A0==offsets[loop2])
                    {
                        printf("%02d|%06X %-42s %s\n",loop2-288,loop,symbols[clue],symbols[loop2]);
                        break;
                    }
                }
            }
            }
        }
        printf("\n");
        for(clue=29;clue<498;clue+=1)
        {
            addr=0;
            for(loop2=288;loop2<370;loop2+=1)
            {
            for(loop=where[clue];loop<where2[clue];loop+=4)
            {
                mips=read32(mem,0x10000000+loop);
                if(op(mips)==JAL)
                {
                    if(target(mips)+0xA771A0==offsets[loop2])
                    {
                        addr+=1;
                        printf("%02d|%06X %-25s %s\n",loop2-288,loop,pointers[clue],symbols[loop2]);
                        break;
                    }
                }
            }
            }
            if(addr>0){printf("count: %d\n",addr);}
        }
        //fclose(text);
        goto pause;
    }
    if(clue>=0x1060)
    {
    for(loop=0;loop<3382;loop+=1)
    {
        if(offsets[loop]==clue)
        {
            printf("%s\n",symbols[loop]);
            break;
        }
    }
    }
    for(loop=0;loop<3381;loop+=1)
    {
        for(loop2=offsets[loop];loop2<offsets[loop+1];loop2+=4)
        {
            mips=read32(mem,0x10000000+loop2);
            if(clue>=0x1060)
            {
            if(op(mips)==JAL)
            {
                if(target(mips)+0xC00==clue||target(mips)+0xA771A0==clue)
                {
                    printf("%06X %s\n",loop2,symbols[loop]);
                    break;
                }
            }
            }
            else
            {
            if(op(mips)==LUI)
            {
                if(imm(mips)==(HALF)clue<<8)
                {
                    printf("%06X %s\n",loop2,symbols[loop]);
                    break;
                }
            }
            }
        }
    }
    for(loop=29;loop<498;loop+=1)
    {
        for(loop2=where[loop];loop2<where2[loop];loop2+=4)
        {
            mips=read32(mem,0x10000000+loop2);
            if(clue>=0x1060)
            {
            if(op(mips)==JAL)
            {
                if(target(mips)+0xC00==clue||target(mips)+0xA771A0==clue)
                {
                    printf("%06X %s\n",loop2,pointers[loop]);
                    break;
                }
            }
            }
            else
            {
            if(op(mips)==LUI)
            {
                if(imm(mips)==(HALF)clue<<8)
                {
                    printf("%06X %s\n",loop2,pointers[loop]);
                    break;
                }
            }
            }
        }
    }
    pause:
    system("pause");
    goto scan;
}

int main(int argc, char **argv)
{
     const char names[32][3]=
     {
          "r0", "at", "v0", "v1",
          "a0", "a1", "a2", "a3",
          "t0", "t1", "t2", "t3",
          "t4", "t5", "t6", "t7",
          "s0", "s1", "s2", "s3",
          "s4", "s5", "s6", "s7",
          "t8", "t9", "k0", "k1",
          "gp", "sp", "fp", "ra"
     };
     BYTE *mem[8], delay=0, depth=0, flag=0, special=0;
     WORD CPU[32]={0},CP0[32]={0},CP1[32]={0},pc=0xa4000040,tick=0,ticks=0,inst;
     unsigned long long DPU[32]={0};
     unsigned long long hi64, lo64, sixtyfour;
     WORD hi, lo;
     unsigned char *output;
     FILE *output2;
     char output3[64];
     FILE *cart=fopen(argv[1],"rb");
     fseek(cart,0,SEEK_END);
     loop=ftell(cart);
     WORD sizes[8]=
     {
          0x800000, /* RAM */
          0x100000, /* REG */
          0x80001c, /* RCP */
          0x0005c0, /* LEO */
          0x400000, /* IPL */
          0x008000, /* STA */
          loop,     /* ROM */
          0x000800  /* PIF */
     };
     for(loop=0;loop<8;loop+=1)
     {
          mem[loop]=(BYTE*)calloc(sizes[loop],1); /* allocate our memory segments and init to zero */
     }
     fseek(cart,0,SEEK_SET);
     fread(mem[ROM],1,sizes[ROM],cart);
     fclose(cart);
     for(loop=0x40;loop<0x1000;loop+=1)
     {
          ticks+=mem[ROM][loop];
          mem[RCP][loop]=mem[ROM][loop];
     }
     switch(ticks)
     {
          case 0x34044: /* MARIO boot code */
               inst=0x00003f3f;
               break;
          case 0x47a81: /* ZELDA boot code */
               inst=0x0000913f;
               break;
     }
     CPU[T3]=pc;
     CPU[S3]=(inst>>19)&1;
     CPU[S4]=1;                          /* tv type */
     CPU[S5]=(inst>>17)&1;
     CPU[S6]=(inst>>8)&0xff;             /* seed */
     CPU[S7]=(inst>>18)&1;
     CPU[SP]=0xa4001ff0;                 /* Stack Pointer */
     CPU[RA]=0xa4001550;                 /* Return Address*/
     CP0[STATUS]=0x34000000;
     CP0[CONFIG]=0x0006e463;
     write32(mem,0xa4001000,0x3c0dbfc0); /* SP IMEM +0 */
     write32(mem,0xa4001004,0x8da807fc); /* SP IMEM +4 */
     write32(mem,0xa4001008,0x25ad07c0); /* SP IMEM +8 */
     write32(mem,0xa400100c,0x31080080); /* SP IMEM +0xC */
     write32(mem,0xa4001010,0x5500fffc); /* SP IMEM +0x10*/
     write32(mem,0xa4001014,0x3c0dbfc0); /* SP IMEM +0x14*/
     write32(mem,0xa4001018,0x8da80024); /* SP IMEM +0x18*/
     write32(mem,0xa400101c,0x3c0bb000); /* SP IMEM +0x1C*/
     write32(mem,0xa4040010,1);          /* SP status */
     write32(mem,0xa4300004,0x02020102); /* MI version*/
     write32(mem,0xbfc007e4,inst);
     inst=read32(mem,0xb0000000);
     write32(mem,0xa4600014,inst&0xff);  /* PI dom1 latency */
     write32(mem,0xa4600018,inst>>8);    /* PI dom1 pulse width */
     write32(mem,0xa460001c,inst>>16);   /* PI dom1 page size */
     write32(mem,0xa4600020,inst>>20);   /* PI dom1 dom1 release*/
     //if(read32(mem,0x1000003c)==0x5A4C450F)
     //{
     //    zelda(mem);
     //    goto skip_emulation;
     //}
     //printf("How many ticks? ");
     //scanf("%d",&ticks);
     CPU[SP]=0x80700000;
     CPU[RA]=0xDEADF00D;
     zelda(mem);
     where2[0]=0;
     write32(mem,0xa4600000,0x400);
     write32(mem,0xa4600004,0x10001000);
     write32(mem,0xa460000c,0x1069F);
     goto skip_emulation;
     emulate:
     while(pc!=0xDEADF00D)
     {
          inst=read32(mem,pc);
          switch(op(inst))
          {
               case SPECIAL:
                    switch(funct(inst))
                    {
                         case SLL:
                              CPU[rd(inst)]=CPU[rt(inst)]<<sa(inst);
                              break;
                         case SRL:
                              CPU[rd(inst)]=CPU[rt(inst)]>>sa(inst);
                              break;
                         case SRA:
                              if((CPU[rt(inst)]&0x80000000)==0x80000000)
                              {
                                  loop=0xFFFFFFFFUL<<(32-sa(inst));
                                  loop|=CPU[rt(inst)]>>sa(inst);
                              }
                              else
                              {
                                  loop=CPU[rt(inst)]>>sa(inst);
                              }
                              CPU[rd(inst)]=loop;
                              break;
                         case SLLV:
                              CPU[rd(inst)]=CPU[rt(inst)]<<(CPU[rs(inst)]&0x1f);
                              break;
                         case SRLV:
                              CPU[rd(inst)]=CPU[rt(inst)]>>(CPU[rs(inst)]&0x1f);
                              break;
                         case JR:
                              if(delay==0)
                              {
                                   delay=1;
                              }
                              else if(delay==3)
                              {
                                   pc=CPU[rs(inst)]-4;
                                   //if(depth==0){depth=1;}
                                   if(pc==CPU[RA]-4)
                                   {
                                        depth-=1;
                                   }
                                   if(rs(inst)!=RA&&pc<0xa0000000)
                                   {
                                       printf("%10d %08x: ",tick+1,pc);
                                       for(loop=0;loop<depth;loop+=1)
                                       {
                                            printf(" ");
                                       }
                                       for(hi=0;hi<3382;hi+=1)
                                       {
                                           if((pc-read32(mem,0xb0000008)+0x1004)==offsets[hi])
                                           {
                                               printf("jumped inside function (%s)\n",symbols[hi]);
                                               break;
                                           }
                                           if(pc+0x80A771A4==offsets[hi])
                                           {
                                               printf("jumped inside function (%s)\n",symbols[hi]);
                                               break;
                                           }
                                       }
                                       if(hi==3382)
                                       {
                                           printf("jumped inside function (%08x)\n",pc+4);
                                       }
                                       //printf("(jr $%s)\n",names[rs(inst)]);
                                       system("pause");
                                       //depth+=1;
                                   }
                                   delay=0;
                              }
                              break;
                         case JALR:
                              if(delay==0)
                              {
                                   CPU[rd(inst)]=pc+8;
                                   delay=1;
                              }
                              else if(delay==3)
                              {
                                   pc=CPU[rs(inst)]-4;
                                   printf("%10d %08x: ",tick+1,pc);
                                   for(loop=0;loop<depth;loop+=1)
                                   {
                                       printf(" ");
                                   }
                                   for(hi=0;hi<3382;hi+=1)
                                   {
                                       if((pc-read32(mem,0xb0000008)+0x1004)==offsets[hi])
                                       {
                                           printf("jumped inside function (%s)\n",symbols[hi]);
                                           break;
                                       }
                                       if(pc+0x80A771A4==offsets[hi])
                                       {
                                           printf("jumped inside function (%s)\n",symbols[hi]);
                                           break;
                                       }
                                   }
                                   if(hi==3382)
                                   {
                                       printf("jumped inside function (%08x)\n",pc+4);
                                   }
                                   system("pause");
                                   depth+=1;
                                   delay=0;
                              }
                              break;
                         case 13:
                              break;
                         case MFHI:
                              CPU[rd(inst)]=hi;
                              DPU[rd(inst)]=hi64;
                              break;
                         case MTHI:
                              hi=CPU[rs(inst)];
                              hi64=DPU[rs(inst)];
                              break;
                         case MFLO:
                              CPU[rd(inst)]=lo;
                              DPU[rd(inst)]=lo64;
                              break;
                         case MTLO:
                              lo=CPU[rs(inst)];
                              lo64=DPU[rs(inst)];
                              break;
                         case MULTU:
                              sixtyfour=CPU[rs(inst)]*CPU[rt(inst)];
                              hi=sixtyfour>>32;
                              lo=sixtyfour&0xffffffff;
                              break;
                         case DIV:
                              hi=(int)CPU[rs(inst)]%(int)CPU[rt(inst)];
                              lo=(int)CPU[rs(inst)]/(int)CPU[rt(inst)];
                              break;
                         case DMULTU:
                              sixtyfour=DPU[rs(inst)]*DPU[rt(inst)];
                              hi64=0;
                              lo64=sixtyfour;
                              break;
                         case DDIV:
                              hi64=(long long)DPU[rs(inst)]%(long long)DPU[rt(inst)];
                              lo64=(long long)DPU[rs(inst)]/(long long)DPU[rt(inst)];
                              break;
                         case DDIVU:
                              hi64=DPU[rs(inst)]%DPU[rt(inst)];
                              lo64=DPU[rs(inst)]/DPU[rt(inst)];
                              break;
                         case ADD:
                              CPU[rd(inst)]=CPU[rs(inst)]+CPU[rt(inst)];
                              break;
                         case ADDU:
                              CPU[rd(inst)]=CPU[rs(inst)]+CPU[rt(inst)];
                              break;
                         case SUBU:
                              CPU[rd(inst)]=CPU[rs(inst)]-CPU[rt(inst)];
                              break;
                         case AND:
                              CPU[rd(inst)]=CPU[rs(inst)]&CPU[rt(inst)];
                              break;
                         case OR:
                              CPU[rd(inst)]=CPU[rs(inst)]|CPU[rt(inst)];
                              break;
                         case XOR:
                              CPU[rd(inst)]=CPU[rs(inst)]^CPU[rt(inst)];
                              break;
                         case NOR:
                              CPU[rd(inst)]=~(CPU[rs(inst)]|CPU[rt(inst)]);
                              break;
                         case SLT:
                              CPU[rd(inst)]=0;
                              if((int)(CPU[rs(inst)])<(int)(CPU[rt(inst)]))
                              {
                                   CPU[rd(inst)]=1;
                              }
                              break;
                         case SLTU:
                              CPU[rd(inst)]=0;
                              if(CPU[rs(inst)]<CPU[rt(inst)])
                              {
                                   CPU[rd(inst)]=1;
                              }
                              break;
                         case DSLL32:
                              DPU[rd(inst)]=DPU[rt(inst)]<<(32+sa(inst));
                              break;
                         case DSRA32:
                              if((DPU[rt(inst)]&0x8000000000000000ULL)==0x8000000000000000ULL)
                              {
                                  sixtyfour=(0xFFFFFFFF00000000ULL)|((0xFFFFFFFF00000000ULL)>>sa(inst));
                                  sixtyfour|=DPU[rt(inst)]>>(32+sa(inst));
                              }
                              else
                              {
                                  sixtyfour=DPU[rt(inst)]>>(32+sa(inst));
                              }
                              DPU[rd(inst)]=sixtyfour;
                              break;
                         default:
                              printf("Unimplemented: %d\n",op(inst));
                              tick=ticks-1;
                              break;
                    }
                    break;
               case REGIMM:
                    switch(rt(inst))
                    {
                         case BLTZ:
                              if(delay==0)
                              {
                                   if((int)CPU[rs(inst)]<0)
                                   {
                                        special=1;
                                   }
                                   delay=1;
                              }
                              else if(delay==3)
                              {
                                   if(special==1){pc+=extend(imm(inst)<<2)-4;special=0;}
                                   pc+=4;
                                   delay=0;
                              }
                              break;
                         case BGEZ:
                              if(delay==0)
                              {
                                   if((int)CPU[rs(inst)]>=0)
                                   {
                                        special=1;
                                   }
                                   delay=1;
                              }
                              else if(delay==3)
                              {
                                   if(special==1){pc+=extend(imm(inst)<<2)-4;special=0;}
                                   pc+=4;
                                   delay=0;
                              }
                              break;
                         case BLTZL:
                              if(delay==0)
                              {
                                   if((int)CPU[rs(inst)]<0)
                                   {
                                        delay=1;
                                   }
                                   else
                                   {
                                        pc+=4;
                                   }
                              }
                              else if(delay==3)
                              {
                                   pc+=extend(imm(inst)<<2);
                                   delay=0;
                              }
                              break;
                         case BGEZL:
                              if(delay==0)
                              {
                                   if((int)CPU[rs(inst)]>=0)
                                   {
                                        delay=1;
                                   }
                                   else
                                   {
                                        pc+=4;
                                   }
                              }
                              else if(delay==3)
                              {
                                   pc+=extend(imm(inst)<<2);
                                   delay=0;
                              }
                              break;
                         case BLTZAL:
                              if(delay==0)
                              {
                                   if((int)CPU[rs(inst)]<0)
                                   {
                                        CPU[RA]=pc+8;
                                        special=1;
                                   }
                                   delay=1;
                              }
                              else if(delay==3)
                              {
                                   if(special==1){pc+=extend(imm(inst)<<2)-4;special=0;}
                                   pc+=4;
                                   delay=0;
                              }
                              break;
                         case BGEZAL:
                              if(delay==0)
                              {
                                   if((int)CPU[rs(inst)]>=0)
                                   {
                                        CPU[RA]=pc+8;
                                        special=1;
                                   }
                                   delay=1;
                              }
                              else if(delay==3)
                              {
                                   if(special==1){pc+=extend(imm(inst)<<2)-4;special=0;}
                                   pc+=4;
                                   delay=0;
                              }
                              break;
                         case BLTZALL:
                              if(delay==0)
                              {
                                   CPU[RA]=pc+8;
                                   if((int)CPU[rs(inst)]<0)
                                   {
                                        delay=1;
                                   }
                                   else
                                   {
                                        pc+=4;
                                   }
                              }
                              else if(delay==3)
                              {
                                   pc+=extend(imm(inst)<<2);
                                   delay=0;
                              }
                              break;
                         case BGEZALL:
                              if(delay==0)
                              {
                                   CPU[RA]=pc+8;
                                   if((int)CPU[rs(inst)]>=0)
                                   {
                                        delay=1;
                                   }
                                   else
                                   {
                                        pc+=4;
                                   }
                              }
                              else if(delay==3)
                              {
                                   pc+=extend(imm(inst)<<2);
                                   delay=0;
                              }
                              break;
                         default:
                              printf("Unimplemented: %d\n",op(inst));
                              tick=ticks-1;
                              break;
                    }
                    break;
               case J:
                    if(delay==0)
                    {
                         delay=1;
                    }
                    else if(delay==3)
                    {
                         pc=((pc&0xf0000000)|target(inst))-4;
                         delay=0;
                    }
                    break;
               case JAL:
                    if(delay==0)
                    {
                         CPU[RA]=pc+8;
                         delay=1;
                    }
                    else if(delay==3)
                    {
                         pc=((pc&0xf0000000)|target(inst))-4;
                         delay=0;
                    }
                    break;
               case BEQ:
                    if(delay==0)
                    {
                         if(CPU[rs(inst)]==CPU[rt(inst)])
                         {
                              special=1;
                         }
                         delay=1;
                    }
                    else if(delay==3)
                    {
                         if(special==1)
                         {
                             pc+=extend(imm(inst)<<2)-4;
                             special=0;
                         }
                         pc+=4;
                         delay=0;
                    }
                    break;
               case BNE:
                    if(delay==0)
                    {
                         if(CPU[rs(inst)]!=CPU[rt(inst)])
                         {
                              special=1;
                         }
                         delay=1;
                    }
                    else if(delay==3)
                    {
                         if(special==1){pc+=extend(imm(inst)<<2)-4;special=0;}
                         pc+=4;
                         delay=0;
                    }
                    break;
               case BLEZ:
                    if(delay==0)
                    {
                         if((int)CPU[rs(inst)]<=0)
                         {
                              special=1;
                         }
                         delay=1;
                    }
                    else if(delay==3)
                    {
                         if(special==1){pc+=extend(imm(inst)<<2)-4;special=0;}
                         pc+=4;
                         delay=0;
                    }
                    break;
               case BGTZ:
                    if(delay==0)
                    {
                         if((int)CPU[rs(inst)]>0)
                         {
                              special=1;
                         }
                         delay=1;
                    }
                    else if(delay==3)
                    {
                         if(special==1){pc+=extend(imm(inst)<<2)-4;special=0;}
                         pc+=4;
                         delay=0;
                    }
                    break;
               case ADDI:
                    CPU[rt(inst)]=CPU[rs(inst)]+extend(imm(inst));
                    break;
               case ADDIU:
                    CPU[rt(inst)]=CPU[rs(inst)]+extend(imm(inst));
                    break;
               case SLTI:
                    CPU[rt(inst)]=0;
                    if((int)(CPU[rs(inst)])<(int)(extend(imm(inst))))
                    {
                         CPU[rt(inst)]=1;
                    }
                    break;
               case SLTIU:
                    CPU[rt(inst)]=0;
                    if(CPU[rs(inst)]<extend(imm(inst)))
                    {
                         CPU[rt(inst)]=1;
                    }
                    break;
               case ANDI:
                    CPU[rt(inst)]=CPU[rs(inst)]&imm(inst);
                    break;
               case ORI:
                    CPU[rt(inst)]=CPU[rs(inst)]|imm(inst);
                    break;
               case XORI:
                    CPU[rt(inst)]=CPU[rs(inst)]^imm(inst);
                    break;
               case LUI:
                    CPU[rt(inst)]=imm(inst)<<16;
                    break;
               case COP0:
                    switch(rs(inst))
                    {
                         case MF:
                              CPU[rt(inst)]=CP0[rd(inst)];
                              break;
                         case MT:
                              CP0[rd(inst)]=CPU[rt(inst)];
                              break;
                         case S:
                              switch(funct(inst))
                              {
                                  case ERET:
                                      CP0[STATUS]&=(WORD)~2;
                                      pc=CP0[EPC]-4;
                                      printf("%10d %08x: ",tick+1,pc);
                                      for(loop=0;loop<depth;loop+=1)
                                      {
                                          printf(" ");
                                      }
                                      if(pc<0xa0000000)
                                      {
                                          for(hi=0;hi<3382;hi+=1)
                                          {
                                              if((pc-read32(mem,0xb0000008)+0x1004)==offsets[hi])
                                              {
                                                  printf("jumped inside function (%s)\n",symbols[hi]);
                                                  break;
                                              }
                                              if(pc+0x80A771A4==offsets[hi])
                                              {
                                                  printf("jumped inside function (%s)\n",symbols[hi]);
                                                  break;
                                              }
                                          }
                                          if(hi==3382)
                                          {
                                              printf("jumped inside function (%08x)\n",pc+4);
                                          }
                                          system("pause");
                                      }
                                      break;
                                  default:
                                      printf("Unimplemented: %d (COP0 funct)\n",funct(inst));
                                      //system("pause");
                                      break;
                              }
                              break;
                         default:
                              printf("Unimplemented: %d (COP0)\n",rs(inst));
                              tick=ticks-1;
                              break;
                    }
                    break;
               case COP1:
                    switch(rs(inst))
                    {
                         case MF:
                         case CF:
                              CPU[rt(inst)]=CP1[rd(inst)];
                              break;
                         case MT:
                         case CT:
                              CP1[rd(inst)]=CPU[rt(inst)];
                              break;
                         default:
                              printf("Unimplemented: %d (COP1)\n",rs(inst));
                              tick=ticks-1;
                              break;
                    }
                    break;
               case BEQL:
                    if(delay==0)
                    {
                         if(CPU[rs(inst)]==CPU[rt(inst)])
                         {
                              delay=1;
                         }
                         else
                         {
                              pc+=4;
                         }
                    }
                    else if(delay==3)
                    {
                         pc+=extend(imm(inst)<<2);
                         delay=0;
                    }
                    break;
               case BNEL:
                    if(delay==0)
                    {
                         if(CPU[rs(inst)]!=CPU[rt(inst)])
                         {
                              delay=1;
                         }
                         else
                         {
                              pc+=4;
                         }
                    }
                    else if(delay==3)
                    {
                         pc+=extend(imm(inst)<<2);
                         delay=0;
                    }
                    break;
               case BLEZL:
                    if(delay==0)
                    {
                         if((int)CPU[rs(inst)]<=0)
                         {
                              delay=1;
                         }
                         else
                         {
                              pc+=4;
                         }
                    }
                    else if(delay==3)
                    {
                         pc+=extend(imm(inst)<<2);
                         delay=0;
                    }
                    break;
               case BGTZL:
                    if(delay==0)
                    {
                         if((int)CPU[rs(inst)]>0)
                         {
                              delay=1;
                         }
                         else
                         {
                              pc+=4;
                         }
                    }
                    else if(delay==3)
                    {
                         pc+=extend(imm(inst)<<2);
                         delay=0;
                    }
                    break;
               case DADDIU:
                    if(sign(imm(inst))==0x8000)
                    {
                        sixtyfour=0xFFFFFFFFFFFF0000ULL|imm(inst);
                    }
                    else
                    {
                        sixtyfour=imm(inst);
                    }
                    DPU[rt(inst)]=DPU[rs(inst)]+sixtyfour;
                    break;
               case LH:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    CPU[rt(inst)]=extend(read32(mem,loop)>>16);
                    break;
               case LW:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    CPU[rt(inst)]=read32(mem,loop);
                    break;
               case LBU:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    CPU[rt(inst)]=read32(mem,loop)>>24;
                    break;
               case LHU:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    CPU[rt(inst)]=read32(mem,loop)>>16;
                    break;
               case SB:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    switch(LUT[(loop>>24)&0x1f])
                    {
                         case REG:
                              mem[REG][loop&0x0fffff]=CPU[rt(inst)]&0xff;
                              break;
                         case PIF:
                              mem[PIF][loop&0x3fffff]=CPU[rt(inst)]&0xff;
                              break;
                         default:
                              mem[LUT[(loop>>24)&0x1f]][loop&0xffffff]=CPU[rt(inst)]&0xff;
                              break;
                    }
                    break;
               case SH:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    switch(LUT[(loop>>24)&0x1f])
                    {
                         case REG:
                              mem[REG][loop&0x0fffff]=(CPU[rt(inst)]>>8)&0xff;
                              mem[REG][(loop&0x0fffff)+1]=CPU[rt(inst)]&0xff;
                              break;
                         case PIF:
                              mem[PIF][loop&0x3fffff]=(CPU[rt(inst)]>>8)&0xff;
                              mem[PIF][(loop&0x3fffff)+1]=CPU[rt(inst)]&0xff;
                              break;
                         default:
                              mem[LUT[(loop>>24)&0x1f]][loop&0xffffff]=(CPU[rt(inst)]>>8)&0xff;
                              mem[LUT[(loop>>24)&0x1f]][(loop&0xffffff)+1]=CPU[rt(inst)]&0xff;
                              break;
                    }
                    break;
               case SW:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    write32(mem,loop,CPU[rt(inst)]);
                    break;
               case CACHE:
                    break;
               case LWC1:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    CP1[rt(inst)]=read32(mem,loop);
                    break;
               case LDC1:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    CP1[rt(inst)]=read32(mem,loop);
                    CP1[rt(inst)+1]=read32(mem,loop+4);
                    break;
               case LD:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    DPU[rt(inst)]=read64(mem,loop);
                    break;
               case SWC1:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    write32(mem,loop,CP1[rt(inst)]);
                    break;
               case SD:
                    loop=CPU[rs(inst)]+extend(imm(inst));
                    write32(mem,loop,DPU[rt(inst)]>>32);
                    write32(mem,loop+4,DPU[rt(inst)]&0xFFFFFFFFUL);
                    break;
               default:
                    printf("Unimplemented: %d\n",op(inst));
                    tick=ticks-1;
                    break;
          }
          pc+=4;
          if(flag==1){tick+=1;}
          if(delay>0)
          {
               delay+=1;
               if(flag==1){tick-=1;}
               if(delay==3)
               {
                    pc-=8;
                    if(flag==1){tick+=1;}
               }
          }
          if(pc==0x80000694)
          {
              pc=0x80000964;
          }
          if(pc==0x80000BFC)
          {
              //depth-=1;
              pc=CPU[RA];
              write32(mem,0xa4600000,CPU[A1]&0x1FFFFFFF);
              write32(mem,0xa4600004,CPU[A0]);
              write32(mem,0xa460000c,CPU[A2]-1);
              //system("pause");
          }
          if(pc==0x80001884)
          {
              pc=0x80001A30;
          }
          if(pc==0x80001A5C||pc==0x80001664)
          {
              depth-=1;
              pc=CPU[RA];
              write32(mem,0xa4600000,CPU[A1]&0x1FFFFFFF);
              write32(mem,0xa4600004,CPU[A2]);
              write32(mem,0xa460000c,CPU[A3]-1);
              system("pause");
          }
          if(pc==0x800017C0)
          {
              depth-=1;
              pc=CPU[RA];
              write32(mem,0xa4600000,CPU[A0]&0x1FFFFFFF);
              write32(mem,0xa4600004,CPU[A1]);
              write32(mem,0xa460000c,CPU[A2]-1);
              system("pause");
          }
          if(pc==0x80002130)
          {
              depth-=1;
              pc=CPU[RA];
              hi=CPU[A0]&0x1FFFFFFF;
              if(hi>=0x1CE60)
              {
              printf("\n");
              if(hi==0x136048)
              {
                  printf("OVL(d):Seg:%08x-%08x Ram:%08x\n",CPU[A1],CPU[A2],CPU[A3]);
                  goto after;
              }
              while(mem[RAM][hi]!=0)
              {
                  printf("%c",mem[RAM][hi]);
                  hi+=1;
              }
              if(mem[RAM][hi-1]!=0xA){printf("\n");}
              after:
              printf("\n");
              system("pause");
              }
          }
          if(pc==0x800C3C20)
          {
              depth-=1;
              pc=CPU[RA];
          }
          if(pc==0x800C6F78)
          {
              pc=0x800C7144;
          }
          if(pc==0x800D6484)
          {
              depth-=1;
              pc=CPU[RA];
          }
          if(pc>=0x800DACC0&&pc<0x800FAF60)
          {
              depth-=1;
              pc=CPU[RA];
          }
     }
     //printf("%s\n",pointers[where2[0]]);
     //system("pause");
     //FILE *dump=fopen("dump","wb");
     hi=read32(mem,0x100116A4+0x10*where2[0])-where[where2[0]];
     output=malloc(hi);
     for(loop=0;loop<hi;loop+=1)
     {
         output[loop]=mem[RAM][0x400000+loop];
     }
     sprintf(output3,"%s",pointers[where2[0]]);
     output2=fopen(output3,"wb");
     fwrite(output,1,hi,output2);
     fclose(output2);
     free(output);
     skip_emulation:
     where2[0]+=1;
     if(read32(mem,0x100116AC+0x10*where2[0])>0)
     {
         CPU[A0]=read32(mem,0x100116A8+0x10*where2[0]);
         CPU[A1]=0x80400000;
         CPU[A2]=read32(mem,0x100116AC+0x10*where2[0])-CPU[A0];
         pc=0x80001B10;
         goto emulate;
     }
     if(where2[0]<1159){goto skip_emulation;}
     for(loop=0;loop<8;loop+=1)
     {
          //if(loop==RAM)
          //{
          //     fwrite(mem[loop],1,sizes[loop],dump);
          //     fclose(dump);
          //}
          free(mem[loop]);
     }
     for(loop=0;loop<32;loop+=1)
     {
          printf("%s: %08x ",names[loop],CPU[loop]);
          if(loop%4==3)
          {
               printf("\n");
          }
     }
     printf("pc: %08x\n",pc);
     system("pause");
     return 0;
}
//printf("f%-2d: %08x ",loop,CP1[loop]);
